#
from .abc_loader import ABLoader, BaseLoader
from .loaders import PDLoader
from .pathfinder import PathFinder

__all__ = ['PathFinder','ABLoader', 'BaseLoader', 'PDLoader']

__version__ = "0.1.4"
__author__ = "Qomp4ss"



import unittest
from file_navigator.abc_loader import BaseLoader

def load_1(path, kwarg1 = None):
    return f"{path} called with kwargs:{kwarg1}"

def load_2(path,  kwarg2 = True):
    return f"{path} called with kwargs:{kwarg2}"
def load_3(path,  kwarg3 = True):
    return f"{path} called with kwargs:{kwarg3}"

class TestBaseLoader(unittest.TestCase):
        
    def test_empty_init(self):
        #empty
        self.assertIsInstance(BaseLoader(), BaseLoader)
        #parametrized
    def test_init_good_args(self):
        arg1 = {load_1: '.txt'}
        arg2 = {load_1: '.txt', load_2: '.json'}
        arg3 = {load_1: '.txt',
                load_2: '.json',
                load_3: ['.xml', '.html']}
        arg4 = {load_1: ['.txt'], 
                load_3: ['.xml', '.html']}
        arg5 = None

        args = (arg1, arg2, arg3, arg4, arg5)
        
        expected1 = {'.txt': load_1}
        expected2 = {'.txt': load_1, '.json': load_2}
        expected3 = {'.txt': load_1,
                     '.json': load_2,
                     '.xml':load_3,
                     '.html':load_3}
        expected4 = {'.txt': load_1,
                     '.xml':load_3,
                     '.html':load_3}
        expected5 = {}

        expected = (expected1, expected2, 
                    expected3, expected4, 
                    expected5)
        
        for i in range(len(args)):
            self.subTest(arg = args[i])
            self.assertIsInstance(BaseLoader(args[i]), BaseLoader)
            self.assertCountEqual(BaseLoader(args[i])._mapp.items(), expected[i].items())

    def test_init_bad_args(self):
        arg1 = [{load_1: '.txt'}]
        arg2 = tuple({load_1: '.txt'})
            
        args = (arg1, arg2)
        for arg in args:
            with self.subTest(arg = arg):
                with self.assertRaises(TypeError):
                    BaseLoader(arg)
                    
    def test_add_functions_bad_arg(self):
        args = [
            {'.txt': load_1},
            {load_1: {'.txt', '.csv'}},
            {load_1: ('.txt', '.csv')},
            {load_1: (1,2,3)}
            ]
        for arg in args:
            with self.subTest(arg = arg):
                with self.assertRaises(TypeError):
                    BaseLoader().add_functions(arg)
    def test_load(self):
        bl = BaseLoader({load_1:'.xlsx',
                         load_2: '.json',
                         load_3: ['.csv', '.txt']})
        mock_dir = r"C:\mock_directory"        
        mock_files = ['Forex.xlsx', 'EURGBP_H4.csv', 'audcad.txt', 'audchf.txt']

        
        mock_paths = ['\\'.join([mock_dir, f]) 
                      if 'xlsx' in f 
                      else '\\'.join(['\\'.join([mock_dir, 'CURR']),f]) if 'csv' in f 
                      else '\\'.join(['\\'.join([mock_dir, f[:3].upper()]), f])
                      for f in mock_files]
        
        args = [
           (mock_paths[0], {'kwarg1': 1, 'kwarg2':'value2', 'kwarg3': 3}),
           (mock_paths[1],  {'kwarg1': 1, 'kwarg2':'value2', 'kwarg3': 3}),
           (mock_paths[2], {'kwarg1': 1, 'kwarg2':'value2', 'kwarg3': 3}),
           (mock_paths[3], {'kwarg1': 1, 'kwarg2':'value2', 'kwarg3': 3})
           ]

        expected = {
            'C:\\mock_directory\\Forex.xlsx': "C:\\mock_directory\\Forex.xlsx called with kwargs:1",
            'C:\\mock_directory\\CURR\\EURGBP_H4.csv': "C:\\mock_directory\\CURR\\EURGBP_H4.csv called with kwargs:3",
            'C:\\mock_directory\\AUD\\audcad.txt': "C:\\mock_directory\\AUD\\audcad.txt called with kwargs:3",
            'C:\\mock_directory\\AUD\\audchf.txt': "C:\\mock_directory\\AUD\\audchf.txt called with kwargs:3"
            }
        
        for arg in args:
            with self.subTest(arg = arg):
                a,b = arg
                self.assertCountEqual(bl.load(a,**b), expected[a])
        
def suite():
    suite = unittest.TestSuite()
    suite.addTest(TestBaseLoader('test_empty_init'))
    suite.addTest(TestBaseLoader('test_init_good_args'))
    suite.addTest(TestBaseLoader('test_init_bad_args'))
    suite.addTest(TestBaseLoader('test_add_functions_bad_arg'))
    suite.addTest(TestBaseLoader('test_load'))
    return suite

if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())


import unittest
from file_navigator.pathfinder import _PathManager
from file_navigator.abc_loader import ABLoader

class Test_PathManager(unittest.TestCase):
       
    def test_empty_init(self):
        with self.assertRaises(ValueError):
           _PathManager([])

    
    def test_select_paths(self):
        mock_dir = r"C:\mock_directory"        
        mock_files = ['Forex.xlsx','EURGBP_H4.csv','EURGBP_M5.csv',
                      'EURJPY_H1.csv','EURJPY_M30.csv', 'EURUSD_M5.csv',
                      'audcad.txt','audchf.txt','audusd.txt',
                      'cadaud.txt','cadeur.txt','cadjpy.txt',
                      'cadpln.txt','cadusd.txt','chfgbp.txt',
                      'chfpln.txt','chfusd.txt','euraud.txt',
                      'eurchf.txt','eurgbp.txt','eurjpy.txt',
                      'eurpln.txt','gbpcad.txt','gbpchf.txt',
                      'gbpeur.txt','gbpjpy.txt','gbppln.txt',
                      'jpyaud.txt','jpypln.txt','nzdusd.txt',
                      'usdcad.txt','usdchf.txt','usdjpy.txt',
                      'usdpln.txt','xaggbp.txt','xauchf.txt',
                      'xaueur.txt','xaugbp.txt']

                
        mock_paths = [
            (mock_dir, f)   if 'xlsx' in f 
            else ('\\'.join([mock_dir, 'CURR']),f) if 'csv' in f 
            else ('\\'.join([mock_dir, f[:3].upper()]), f)
            for f in mock_files
            ]
        
        args = [
            ('J.*|X.*',  'regex'),
            ('*',  'glob'),
            ('N', 'isin'),
            ('XAG|NZD', 'regex'),
            (r'C:\mock_directory\CURR', 'eq')
            ]
        expected = {
            ('J.*|X.*',  'regex'): ['jpyaud.txt', 'jpypln.txt', 'xaggbp.txt',
                                    'xauchf.txt', 'xaueur.txt', 'xaugbp.txt'],
            ('*', 'glob'): mock_files,
            ('N', 'isin'): ['nzdusd.txt'],
            ('XAG|NZD', 'regex'): ['xaggbp.txt', 'nzdusd.txt'],
            (r'C:\mock_directory\CURR', 'eq'): ['EURGBP_H4.csv', 'EURGBP_M5.csv',
                                                'EURJPY_H1.csv','EURJPY_M30.csv', 
                                                'EURUSD_M5.csv']
                }
        
        for arg in args:
            with self.subTest(arg = arg):
                result = [p[0] for p in _PathManager(mock_paths).select_paths(*arg).paths]
                self.assertCountEqual(result, expected[arg])
                with self.assertRaises(ValueError):
                    _PathManager(mock_paths).select_paths(('XXXXXXX',  'eq'))
        

    
    def test_groupby(self):
        mock_dir = r"C:\mock_directory"        
        mock_files = ['Forex.xlsx','EURGBP_H4.csv','EURGBP_M5.csv',
                      'EURJPY_H1.csv','EURJPY_M30.csv', 'EURUSD_M5.csv',
                      'audcad.txt','audchf.txt','audusd.txt',
                      'cadaud.txt','cadeur.txt','cadjpy.txt',
                      'cadpln.txt','cadusd.txt','chfgbp.txt',
                      'chfpln.txt','chfusd.txt','euraud.txt',
                      'eurchf.txt','eurgbp.txt','eurjpy.txt',
                      'eurpln.txt','gbpcad.txt','gbpchf.txt',
                      'gbpeur.txt','gbpjpy.txt','gbppln.txt',
                      'jpyaud.txt','jpypln.txt','nzdusd.txt',
                      'usdcad.txt','usdchf.txt','usdjpy.txt',
                      'usdpln.txt','xaggbp.txt','xauchf.txt',
                      'xaueur.txt','xaugbp.txt']

                
        mock_paths = [
            (mock_dir, f)   if 'xlsx' in f 
            else ('\\'.join([mock_dir, 'CURR']),f) if 'csv' in f 
            else ('\\'.join([mock_dir, f[:3].upper()]), f)
            for f in mock_files
            ]
        args = [('ext',None),
                ('path',None),
                ('name',None),
                ('ext','.txt','eq'), 
                ('path', '*AG','glob'), 
                ('name', r'^eur.*|^chf.*', 'regex')]
        expected = {
            ('ext',None): {'.xlsx':_PathManager(('C:\\mock_directory', 'Forex.xlsx')).paths,
                    '.csv':_PathManager([('C:\\mock_directory\\CURR', 'EURGBP_H4.csv'),
                                       ('C:\\mock_directory\\CURR', 'EURGBP_M5.csv'),
                                       ('C:\\mock_directory\\CURR', 'EURJPY_H1.csv'),
                                       ('C:\\mock_directory\\CURR', 'EURJPY_M30.csv'),
                                       ('C:\\mock_directory\\CURR', 'EURUSD_M5.csv')]
                                       ).paths,
                    '.txt':_PathManager([('C:\\mock_directory\\AUD', 'audcad.txt'),
                                        ('C:\\mock_directory\\AUD', 'audchf.txt'),
                                        ('C:\\mock_directory\\AUD', 'audusd.txt'),
                                        ('C:\\mock_directory\\CAD', 'cadaud.txt'),
                                        ('C:\\mock_directory\\CAD', 'cadeur.txt'),
                                        ('C:\\mock_directory\\CAD', 'cadjpy.txt'),
                                        ('C:\\mock_directory\\CAD', 'cadpln.txt'),
                                        ('C:\\mock_directory\\CAD', 'cadusd.txt'),
                                        ('C:\\mock_directory\\CHF', 'chfgbp.txt'),
                                        ('C:\\mock_directory\\CHF', 'chfpln.txt'),
                                        ('C:\\mock_directory\\CHF', 'chfusd.txt'),
                                        ('C:\\mock_directory\\EUR', 'euraud.txt'),
                                        ('C:\\mock_directory\\EUR', 'eurchf.txt'),
                                        ('C:\\mock_directory\\EUR', 'eurgbp.txt'),
                                        ('C:\\mock_directory\\EUR', 'eurjpy.txt'),
                                        ('C:\\mock_directory\\EUR', 'eurpln.txt'),
                                        ('C:\\mock_directory\\GBP', 'gbpcad.txt'),
                                        ('C:\\mock_directory\\GBP', 'gbpchf.txt'),
                                        ('C:\\mock_directory\\GBP', 'gbpeur.txt'),
                                        ('C:\\mock_directory\\GBP', 'gbpjpy.txt'),
                                        ('C:\\mock_directory\\GBP', 'gbppln.txt'),
                                        ('C:\\mock_directory\\JPY', 'jpyaud.txt'),
                                        ('C:\\mock_directory\\JPY', 'jpypln.txt'),
                                        ('C:\\mock_directory\\NZD', 'nzdusd.txt'),
                                        ('C:\\mock_directory\\USD', 'usdcad.txt'),
                                        ('C:\\mock_directory\\USD', 'usdchf.txt'),
                                        ('C:\\mock_directory\\USD', 'usdjpy.txt'),
                                        ('C:\\mock_directory\\USD', 'usdpln.txt'),
                                        ('C:\\mock_directory\\XAG', 'xaggbp.txt'),
                                        ('C:\\mock_directory\\XAU', 'xauchf.txt'),
                                        ('C:\\mock_directory\\XAU', 'xaueur.txt'),
                                        ('C:\\mock_directory\\XAU', 'xaugbp.txt')]
                                        ).paths
                    },
            ('path',None):{
                'C:\\mock_directory': _PathManager(
                    ('C:\\mock_directory', 'Forex.xlsx')
                    ).paths,
                'C:\\mock_directory\\CURR': _PathManager(
                    [('C:\\mock_directory\\CURR', 'EURGBP_H4.csv'),
                    ('C:\\mock_directory\\CURR', 'EURGBP_M5.csv'),
                    ('C:\\mock_directory\\CURR', 'EURJPY_H1.csv'),
                    ('C:\\mock_directory\\CURR', 'EURJPY_M30.csv'),
                    ('C:\\mock_directory\\CURR', 'EURUSD_M5.csv')]
                    ).paths,
                'C:\\mock_directory\\AUD': _PathManager(
                    [('C:\\mock_directory\\AUD', 'audcad.txt'),
                    ('C:\\mock_directory\\AUD', 'audchf.txt'),
                    ('C:\\mock_directory\\AUD', 'audusd.txt')]
                    ).paths,
                'C:\\mock_directory\\CAD': _PathManager(
                    [('C:\\mock_directory\\CAD', 'cadaud.txt'),
                    ('C:\\mock_directory\\CAD', 'cadeur.txt'),
                    ('C:\\mock_directory\\CAD', 'cadjpy.txt'),
                    ('C:\\mock_directory\\CAD', 'cadpln.txt'),
                    ('C:\\mock_directory\\CAD', 'cadusd.txt')]
                    ).paths,
                'C:\\mock_directory\\CHF': _PathManager(
                    [('C:\\mock_directory\\CHF', 'chfgbp.txt'),
                    ('C:\\mock_directory\\CHF', 'chfpln.txt'),
                    ('C:\\mock_directory\\CHF', 'chfusd.txt')]
                    ).paths,
                'C:\\mock_directory\\EUR': _PathManager(
                    [('C:\\mock_directory\\EUR', 'euraud.txt'),
                    ('C:\\mock_directory\\EUR', 'eurchf.txt'),
                    ('C:\\mock_directory\\EUR', 'eurgbp.txt'),
                    ('C:\\mock_directory\\EUR', 'eurjpy.txt'),
                    ('C:\\mock_directory\\EUR', 'eurpln.txt')]
                    ).paths,
                'C:\\mock_directory\\GBP': _PathManager(
                    [('C:\\mock_directory\\GBP', 'gbpcad.txt'),
                    ('C:\\mock_directory\\GBP', 'gbpchf.txt'),
                    ('C:\\mock_directory\\GBP', 'gbpeur.txt'),
                    ('C:\\mock_directory\\GBP', 'gbpjpy.txt'),
                    ('C:\\mock_directory\\GBP', 'gbppln.txt')]
                    ).paths,
                'C:\\mock_directory\\JPY': _PathManager(
                    [('C:\\mock_directory\\JPY', 'jpyaud.txt'),
                    ('C:\\mock_directory\\JPY', 'jpypln.txt')]
                    ).paths,
                'C:\\mock_directory\\NZD': _PathManager(
                    [('C:\\mock_directory\\NZD', 'nzdusd.txt')]
                    ).paths,
                'C:\\mock_directory\\USD': _PathManager(
                    [('C:\\mock_directory\\USD', 'usdcad.txt'),
                    ('C:\\mock_directory\\USD', 'usdchf.txt'),
                    ('C:\\mock_directory\\USD', 'usdjpy.txt'),
                    ('C:\\mock_directory\\USD', 'usdpln.txt')]
                    ).paths,
                'C:\\mock_directory\\XAG': _PathManager(
                    ('C:\\mock_directory\\XAG', 'xaggbp.txt')
                    ).paths,
                'C:\\mock_directory\\XAU': _PathManager(
                    [('C:\\mock_directory\\XAU', 'xauchf.txt'),
                    ('C:\\mock_directory\\XAU', 'xaueur.txt'),
                    ('C:\\mock_directory\\XAU', 'xaugbp.txt')]
                    ).paths
                },
            ('name',None):{
                'Forex': _PathManager(('C:\\mock_directory', 'Forex.xlsx')).paths,
                'EURGBP_H4': _PathManager(('C:\\mock_directory\\CURR', 'EURGBP_H4.csv')).paths,
                'EURGBP_M5': _PathManager(('C:\\mock_directory\\CURR', 'EURGBP_M5.csv')).paths,
                'EURJPY_H1': _PathManager(('C:\\mock_directory\\CURR', 'EURJPY_H1.csv')).paths,
                'EURJPY_M30': _PathManager(('C:\\mock_directory\\CURR', 'EURJPY_M30.csv')).paths,
                'EURUSD_M5': _PathManager(('C:\\mock_directory\\CURR', 'EURUSD_M5.csv')).paths,
                'audcad': _PathManager(('C:\\mock_directory\\AUD', 'audcad.txt')).paths,
                'audchf': _PathManager(('C:\\mock_directory\\AUD', 'audchf.txt')).paths,
                'audusd': _PathManager(('C:\\mock_directory\\AUD', 'audusd.txt')).paths,
                'cadaud': _PathManager(('C:\\mock_directory\\CAD', 'cadaud.txt')).paths,
                'cadeur': _PathManager(('C:\\mock_directory\\CAD', 'cadeur.txt')).paths,
                'cadjpy': _PathManager(('C:\\mock_directory\\CAD', 'cadjpy.txt')).paths,
                'cadpln': _PathManager(('C:\\mock_directory\\CAD', 'cadpln.txt')).paths,
                'cadusd': _PathManager(('C:\\mock_directory\\CAD', 'cadusd.txt')).paths,
                'chfgbp': _PathManager(('C:\\mock_directory\\CHF', 'chfgbp.txt')).paths,
                'chfpln': _PathManager(('C:\\mock_directory\\CHF', 'chfpln.txt')).paths,
                'chfusd': _PathManager(('C:\\mock_directory\\CHF', 'chfusd.txt')).paths,
                'euraud': _PathManager(('C:\\mock_directory\\EUR', 'euraud.txt')).paths,
                'eurchf': _PathManager(('C:\\mock_directory\\EUR', 'eurchf.txt')).paths,
                'eurgbp': _PathManager(('C:\\mock_directory\\EUR', 'eurgbp.txt')).paths,
                'eurjpy': _PathManager(('C:\\mock_directory\\EUR', 'eurjpy.txt')).paths,
                'eurpln': _PathManager(('C:\\mock_directory\\EUR', 'eurpln.txt')).paths,
                'gbpcad': _PathManager(('C:\\mock_directory\\GBP', 'gbpcad.txt')).paths,
                'gbpchf': _PathManager(('C:\\mock_directory\\GBP', 'gbpchf.txt')).paths,
                'gbpeur': _PathManager(('C:\\mock_directory\\GBP', 'gbpeur.txt')).paths,
                'gbpjpy': _PathManager(('C:\\mock_directory\\GBP', 'gbpjpy.txt')).paths,
                'gbppln': _PathManager(('C:\\mock_directory\\GBP', 'gbppln.txt')).paths,
                'jpyaud': _PathManager(('C:\\mock_directory\\JPY', 'jpyaud.txt')).paths,
                'jpypln': _PathManager(('C:\\mock_directory\\JPY', 'jpypln.txt')).paths,
                'nzdusd': _PathManager(('C:\\mock_directory\\NZD', 'nzdusd.txt')).paths,
                'usdcad': _PathManager(('C:\\mock_directory\\USD', 'usdcad.txt')).paths,
                'usdchf': _PathManager(('C:\\mock_directory\\USD', 'usdchf.txt')).paths,
                'usdjpy': _PathManager(('C:\\mock_directory\\USD', 'usdjpy.txt')).paths,
                'usdpln': _PathManager(('C:\\mock_directory\\USD', 'usdpln.txt')).paths,
                'xaggbp': _PathManager(('C:\\mock_directory\\XAG', 'xaggbp.txt')).paths,
                'xauchf': _PathManager(('C:\\mock_directory\\XAU', 'xauchf.txt')).paths,
                'xaueur': _PathManager(('C:\\mock_directory\\XAU', 'xaueur.txt')).paths,
                'xaugbp': _PathManager(('C:\\mock_directory\\XAU', 'xaugbp.txt')).paths
                },
            ('ext','.txt','eq'): {
                False:_PathManager([('C:\\mock_directory', 'Forex.xlsx'),
                                    ('C:\\mock_directory\\CURR', 'EURGBP_H4.csv'),
                                    ('C:\\mock_directory\\CURR', 'EURGBP_M5.csv'),
                                    ('C:\\mock_directory\\CURR', 'EURJPY_H1.csv'),
                                    ('C:\\mock_directory\\CURR', 'EURJPY_M30.csv'),
                                    ('C:\\mock_directory\\CURR', 'EURUSD_M5.csv')
                                    ]).paths,
                True:_PathManager([('C:\\mock_directory\\AUD', 'audcad.txt'),
                                   ('C:\\mock_directory\\AUD', 'audchf.txt'),
                                   ('C:\\mock_directory\\AUD', 'audusd.txt'),
                                   ('C:\\mock_directory\\CAD', 'cadaud.txt'),
                                   ('C:\\mock_directory\\CAD', 'cadeur.txt'),
                                   ('C:\\mock_directory\\CAD', 'cadjpy.txt'),
                                   ('C:\\mock_directory\\CAD', 'cadpln.txt'),
                                   ('C:\\mock_directory\\CAD', 'cadusd.txt'),
                                   ('C:\\mock_directory\\CHF', 'chfgbp.txt'),
                                   ('C:\\mock_directory\\CHF', 'chfpln.txt'),
                                   ('C:\\mock_directory\\CHF', 'chfusd.txt'),
                                   ('C:\\mock_directory\\EUR', 'euraud.txt'),
                                   ('C:\\mock_directory\\EUR', 'eurchf.txt'),
                                   ('C:\\mock_directory\\EUR', 'eurgbp.txt'),
                                   ('C:\\mock_directory\\EUR', 'eurjpy.txt'),
                                   ('C:\\mock_directory\\EUR', 'eurpln.txt'),
                                   ('C:\\mock_directory\\GBP', 'gbpcad.txt'),
                                   ('C:\\mock_directory\\GBP', 'gbpchf.txt'),
                                   ('C:\\mock_directory\\GBP', 'gbpeur.txt'),
                                   ('C:\\mock_directory\\GBP', 'gbpjpy.txt'),
                                   ('C:\\mock_directory\\GBP', 'gbppln.txt'),
                                   ('C:\\mock_directory\\JPY', 'jpyaud.txt'),
                                   ('C:\\mock_directory\\JPY', 'jpypln.txt'),
                                   ('C:\\mock_directory\\NZD', 'nzdusd.txt'),
                                   ('C:\\mock_directory\\USD', 'usdcad.txt'),
                                   ('C:\\mock_directory\\USD', 'usdchf.txt'),
                                   ('C:\\mock_directory\\USD', 'usdjpy.txt'),
                                   ('C:\\mock_directory\\USD', 'usdpln.txt'),
                                   ('C:\\mock_directory\\XAG', 'xaggbp.txt'),
                                   ('C:\\mock_directory\\XAU', 'xauchf.txt'),
                                   ('C:\\mock_directory\\XAU', 'xaueur.txt'),
                                   ('C:\\mock_directory\\XAU', 'xaugbp.txt')
                                    ]).paths
                }, 
            ('path', '*AG','glob'): {
                False:_PathManager([('C:\\mock_directory', 'Forex.xlsx'),
                                    ('C:\\mock_directory\\CURR', 'EURGBP_H4.csv'),
                                    ('C:\\mock_directory\\CURR', 'EURGBP_M5.csv'),
                                    ('C:\\mock_directory\\CURR', 'EURJPY_H1.csv'),
                                    ('C:\\mock_directory\\CURR', 'EURJPY_M30.csv'),
                                    ('C:\\mock_directory\\CURR', 'EURUSD_M5.csv'),
                                    ('C:\\mock_directory\\AUD', 'audcad.txt'),
                                    ('C:\\mock_directory\\AUD', 'audchf.txt'),
                                    ('C:\\mock_directory\\AUD', 'audusd.txt'),
                                    ('C:\\mock_directory\\CAD', 'cadaud.txt'),
                                    ('C:\\mock_directory\\CAD', 'cadeur.txt'),
                                    ('C:\\mock_directory\\CAD', 'cadjpy.txt'),
                                    ('C:\\mock_directory\\CAD', 'cadpln.txt'),
                                    ('C:\\mock_directory\\CAD', 'cadusd.txt'),
                                    ('C:\\mock_directory\\CHF', 'chfgbp.txt'),
                                    ('C:\\mock_directory\\CHF', 'chfpln.txt'),
                                    ('C:\\mock_directory\\CHF', 'chfusd.txt'),
                                    ('C:\\mock_directory\\EUR', 'euraud.txt'),
                                    ('C:\\mock_directory\\EUR', 'eurchf.txt'),
                                    ('C:\\mock_directory\\EUR', 'eurgbp.txt'),
                                    ('C:\\mock_directory\\EUR', 'eurjpy.txt'),
                                    ('C:\\mock_directory\\EUR', 'eurpln.txt'),
                                    ('C:\\mock_directory\\GBP', 'gbpcad.txt'),
                                    ('C:\\mock_directory\\GBP', 'gbpchf.txt'),
                                    ('C:\\mock_directory\\GBP', 'gbpeur.txt'),
                                    ('C:\\mock_directory\\GBP', 'gbpjpy.txt'),
                                    ('C:\\mock_directory\\GBP', 'gbppln.txt'),
                                    ('C:\\mock_directory\\JPY', 'jpyaud.txt'),
                                    ('C:\\mock_directory\\JPY', 'jpypln.txt'),
                                    ('C:\\mock_directory\\NZD', 'nzdusd.txt'),
                                    ('C:\\mock_directory\\USD', 'usdcad.txt'),
                                    ('C:\\mock_directory\\USD', 'usdchf.txt'),
                                    ('C:\\mock_directory\\USD', 'usdjpy.txt'),
                                    ('C:\\mock_directory\\USD', 'usdpln.txt'),
                                    ('C:\\mock_directory\\XAU', 'xauchf.txt'),
                                    ('C:\\mock_directory\\XAU', 'xaueur.txt'),
                                    ('C:\\mock_directory\\XAU', 'xaugbp.txt')
                                    ]).paths,
                True:_PathManager([('C:\\mock_directory\\XAG', 'xaggbp.txt')]).paths
                }, 
            ('name', r'^eur.*|^chf.*', 'regex'): {
                False:_PathManager([('C:\\mock_directory', 'Forex.xlsx'),
                                    ('C:\\mock_directory\\CURR', 'EURGBP_H4.csv'),
                                    ('C:\\mock_directory\\CURR', 'EURGBP_M5.csv'),
                                    ('C:\\mock_directory\\CURR', 'EURJPY_H1.csv'),
                                    ('C:\\mock_directory\\CURR', 'EURJPY_M30.csv'),
                                    ('C:\\mock_directory\\CURR', 'EURUSD_M5.csv'),
                                    ('C:\\mock_directory\\AUD', 'audcad.txt'),
                                    ('C:\\mock_directory\\AUD', 'audchf.txt'),
                                    ('C:\\mock_directory\\AUD', 'audusd.txt'),
                                    ('C:\\mock_directory\\CAD', 'cadaud.txt'),
                                    ('C:\\mock_directory\\CAD', 'cadeur.txt'),
                                    ('C:\\mock_directory\\CAD', 'cadjpy.txt'),
                                    ('C:\\mock_directory\\CAD', 'cadpln.txt'),
                                    ('C:\\mock_directory\\CAD', 'cadusd.txt'),
                                    ('C:\\mock_directory\\GBP', 'gbpcad.txt'),
                                    ('C:\\mock_directory\\GBP', 'gbpchf.txt'),
                                    ('C:\\mock_directory\\GBP', 'gbpeur.txt'),
                                    ('C:\\mock_directory\\GBP', 'gbpjpy.txt'),
                                    ('C:\\mock_directory\\GBP', 'gbppln.txt'),
                                    ('C:\\mock_directory\\JPY', 'jpyaud.txt'),
                                    ('C:\\mock_directory\\JPY', 'jpypln.txt'),
                                    ('C:\\mock_directory\\NZD', 'nzdusd.txt'),
                                    ('C:\\mock_directory\\USD', 'usdcad.txt'),
                                    ('C:\\mock_directory\\USD', 'usdchf.txt'),
                                    ('C:\\mock_directory\\USD', 'usdjpy.txt'),
                                    ('C:\\mock_directory\\USD', 'usdpln.txt'),
                                    ('C:\\mock_directory\\XAG', 'xaggbp.txt'),
                                    ('C:\\mock_directory\\XAU', 'xauchf.txt'),
                                    ('C:\\mock_directory\\XAU', 'xaueur.txt'),
                                    ('C:\\mock_directory\\XAU', 'xaugbp.txt')
                                    ]).paths,
                True:_PathManager([('C:\\mock_directory\\CHF', 'chfgbp.txt'),
                                   ('C:\\mock_directory\\CHF', 'chfpln.txt'),
                                   ('C:\\mock_directory\\CHF', 'chfusd.txt'),
                                   ('C:\\mock_directory\\EUR', 'euraud.txt'),
                                   ('C:\\mock_directory\\EUR', 'eurchf.txt'),
                                   ('C:\\mock_directory\\EUR', 'eurgbp.txt'),
                                   ('C:\\mock_directory\\EUR', 'eurjpy.txt'),
                                   ('C:\\mock_directory\\EUR', 'eurpln.txt')
                                   ]).paths
                }
            }
        
        for arg in args:
            with self.subTest(arg = arg):
                result = {k: v.paths for k, v in _PathManager(mock_paths).groupby(*arg).items()}
                self.assertCountEqual(result, expected[arg])
                self.assertCountEqual(result.values(), expected[arg].values())
        
               
    def test_load(self):
        
        class MockLoader(ABLoader):
            
            def load(self, path, **kwargs):
                return f"{path} with {kwargs}"
            
        mock_dir = r"C:\mock_directory"        
        mock_files = ['Forex.xlsx', 'EURGBP_H4.csv', 'audcad.txt', 'audchf.txt']

                
        mock_paths = [
            (mock_dir, f)   if 'xlsx' in f 
            else ('\\'.join([mock_dir, 'CURR']),f) if 'csv' in f 
            else ('\\'.join([mock_dir, f[:3].upper()]), f)
            for f in mock_files
            ]
            
        kwargs = {'key1': 'value1'}
        expected = [
                    f"{p[0]} with {p[1]}" for p in 
                    [('\\'.join(p), kwargs)for p in mock_paths]
                    ]
        pm = _PathManager(mock_paths)
        self.assertCountEqual(pm.load(MockLoader(), **kwargs), expected)
        
    def test_load_bad_loader(self):
        
        class MockLoader:
            
            def load(self, path):
                return "Im not following ABLoader interface!"

        pm = _PathManager([(r"C:\mock_directory", 'Forex.xlsx')])
        kwargs = {'key1': 'value1'}
        with self.assertRaises(TypeError):
            pm.load(MockLoader(), **kwargs)
        

    
def suite():
    suite = unittest.TestSuite()
    suite.addTest(Test_PathManager('test_empty_init'))
    suite.addTest(Test_PathManager('test_select_paths'))
    suite.addTest(Test_PathManager('test_groupby'))
    suite.addTest(Test_PathManager('test_load'))
    suite.addTest(Test_PathManager('test_load_bad_loader'))
    return suite

if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())


import re
from pathlib import Path

def eq(string, pattern):
    """
    Equality matching function.
    
    Parameters
    ----------
    string: str
        String with which pattern will be compared.
    pattern: str
        String pattern to be matched.
    
    Returns
    -------
    bool
        True/False if string and a pattern are equal.
    """
    return string == pattern
          
def isin(string, pattern):
    """
    Inclusion matching function.
    
    Parameters
    ----------
    string: str
        String in which pattern will be searched.
    pattern: str
        String pattern to be searched.
    
    Returns
    -------
    bool
        True/False if pattern is in the string.
    """
    return pattern in string
          
def regex(string, pattern):
    """
    Regex matching function supporting standard strings and file paths.
    
    Parameters
    ----------
    string: str
        String in which regex pattern will be searched.
    pattern: str
        String regex pattern to be searched.
    
    Returns
    -------
    bool
        True/False if regex pattern was found in the string.
    """
    return True if re.search(pattern, string) else False

def glob(string, pattern):
    """
    Glob matching function supporting standard strings and file paths.
    
    Parameters
    ----------
    string: str
        String in which glob pattern will be searched.
    pattern: str
        String glob pattern to be searched.
    
    Returns
    -------
    bool
        True/False if glob pattern was found in the string.
    """
    return Path(string).match(pattern)



from .abc_loader import BaseLoader
import pandas as pd

PDLoader = BaseLoader({pd.read_csv: ['.csv', '.txt'],
                       pd.read_excel: ['.xlsx', '.xls'],
                       pd.read_feather: '.feather',
                       pd.read_hdf:['.h5', '.hdf5'],
                       pd.read_json: '.json',
                       pd.read_parquet:'.gzip',
                       pd.read_pickle: '.pkl',
                       pd.read_sas: '.sas7bdat',
                       pd.read_stata: '.dta',
                       pd.read_html: '.html'})


from abc import ABC, abstractmethod
import inspect
from pathlib import Path
from functools import lru_cache

class ABLoader(ABC):
    """
    Loader Abstract Base Class
    
    Interface which all Loader objects must implement, to ensure that 
    dependency injection in _PathManager.load method will work as expected.
    """
    @abstractmethod
    def load(self, path, **kwargs):
        """
        Abstract function for loading data using file path.
        
        Parameters
        ----------
        path: path-like object.
        kwargs: dict
        """
        raise NotImplementedError("This has to be implemented")

class BaseLoader(ABLoader):
    """
    Simple Loader Factory
    
    Class that implements ABLoader interface allowing to dynamically create loader objects, 
    based on a dictionary with loading functions and extension_type(s) pairs.
    
    Parameters:
        func_ftype_dict (Dict[callable : str | list][default=None]): Dictionary 
            with loading functions and extension_type(s) pairs.
    
    Attributes:
        _mapp (dict): Empty dict to which loading functions and extension_type(s)
            pairs will be added.
    
    Methods:
        add_functions (func_ftype_dict: Dict[callable : str | list]]): Function
            to add at least one loading function and extension_type(s) entry.
        load (path: str, kwargs: dict): Function for loading data specified by 
            a file path, distributing matching key-value arguments to all of the 
            available loader functions.            
    """
    def __init__(self, func_ftype_dict = None):
        self._mapp = {}
        if func_ftype_dict is not None:
            self.add_functions(func_ftype_dict)
        
    def _add(self, func, file_type):
        """
        Private function for adding single mapping of loader function and file type.
        
        Function that adds and validates a single entry of a loader function 
        and file type mapping.
        
        Parameters
        ----------
        func: Callable
            Loading function.
        file_type: str | list
            String or list of strings representing file types (extensions).
        
        Returns
        -------
        None
        """
        if not callable(func):
            raise TypeError(f"{func} is not callable")
            
        if not isinstance(file_type, (str, list)):
            raise TypeError(f"{file_type} must be passed as either string or list")
                
        if isinstance(file_type, list):
            for ft in set(file_type):
                if not isinstance(ft, str):
                    raise TypeError(f"{ft} must be passed as a string")
                self._mapp[ft] = func
        else:
            self._mapp[file_type] = func
        
    def add_functions(self, func_ftype_dict):
        """
        Function for adding multiple loader function and file type mappings.
        
        Function that allows adding multiple loader function and file type 
        mappings, by iteratively calling _add method.
        
        Parameters
        ----------
        func_ftype_dict: Dict[callable : str | list]
            Dictionary with loading functions and extension_type(s) pairs.
        
        Returns
        -------
        None
        """
        if not isinstance(func_ftype_dict, dict):
            raise TypeError('func_ftype_dict must be of dictionary type')
        for k, v in func_ftype_dict.items():
            self._add(k, v)
    
    def _inspect(self, function):
        """
        Private function for returning all of the parameters accepted by a loader function.
        
        Parameters
        ----------
        function: Callable
            Function whose parameters have to be retrieved.
        
        Returns
        -------
        odict_keys
            Collection of the parameters accepted by the function.
        """
        return inspect.signature(function).parameters.keys()
    
    @lru_cache(maxsize=None)            
    def load(self, path, **kwargs):
        """
        Implementation of ABLoader interface.
        
        Function implementing ABLoader interface, delegating loading data 
        operations to a loader function based on the pre-defined file type,
        from a file specified by a path-like string, distributing matching key-value 
        arguments to all of the available loader functions.
        
        Parameters
        ----------
        path: str
            Path-like string pointing to an existing file.
        kwargs: dict
            Key, value arguments to be distributed to loader functions.
        
        Returns
        -------
        obj
            Object loaded by the delegate function.
        """
        f = self._mapp[Path(path).suffix]
        return f(path, **{k:v for k, v in kwargs.items() if k in self._inspect(f)})
    


import unittest
from file_navigator import PathFinder
from pathlib import Path
from unittest.mock import patch
from itertools import chain

class TestPathFinder(unittest.TestCase):
    
    def test_empty_init(self):
        self.assertIsInstance(PathFinder(), PathFinder)

    def test_init_good_arg(self):
        arg1 = {r"C:\\": 1}
        arg2 = {r"C:\\": 0}
        arg3 = {r"C:\\": True}
        arg4 = {r"C:\\": False}
        
        args = (arg1, arg2, arg3, arg4)
        
        for arg in args:
            self.subTest(arg = arg)
            self.assertIsInstance(PathFinder(arg), PathFinder)

    @unittest.expectedFailure
    def test_init_bad_arg_1(self):
        arg = [1, "2"]
        
        self.assertIsInstance(PathFinder(arg), PathFinder)

    @unittest.expectedFailure
    def test_init_bad_arg_2(self):
        arg =  {1: True}
        
        self.assertIsInstance(PathFinder(arg), PathFinder)

    @unittest.expectedFailure
    def test_init_bad_arg_3(self):
        arg = {r"C:\\": 12}
        
        self.assertIsInstance(PathFinder(arg), PathFinder)

    @unittest.expectedFailure
    def test_init_bad_arg_4(self):
        arg = {r"C:\\": "True"}

        
        self.assertIsInstance(PathFinder(arg), PathFinder)


    @unittest.expectedFailure
    def test_init_bad_arg_5(self):
        arg = {r"ABC": False}
        
        self.assertIsInstance(PathFinder(arg), PathFinder)
        
    @unittest.expectedFailure
    def test_init_bad_arg_6(self):
        arg = {r"ABC": None}
        
        self.assertIsInstance(PathFinder(arg), PathFinder)
        
    @patch('file_navigator.pathfinder.os.path.isdir')  
    def test_add_dirs(self, mock_isdir):
        directories =  {"/folder1": True,
                        "/folder2/subfolderA": False,
                        "/folder3/subfolderB/subsubfolderC": True}
        
        mock_isdir.side_effect = lambda path: True if path in directories else False
        pf = PathFinder()
        pf.add_dirs(directories)
        self.assertEqual(directories, pf.directories)
    
    @patch('file_navigator.pathfinder.os.path.isdir')     
    def test_add_dirs_bad_arg(self, mock_isdir):
        
        directories = ['/root/subfolder1', '/root/subfolder2',
                       '/root/subfolder3', '/root']
        
        arg1 = ({"/root/subfolder1": False,
                "/root/subfolder2": True,
                "/root/subfolder3": False},
                {"/root": True},)
        
        arg2= ({"/root": True},
                {"/root/subfolder3": False})

        

        args = (arg1, arg2)
        mock_isdir.side_effect = lambda path: True if path in directories else False
        
        for arg in args:
            with self.subTest(arg = arg):
                pf = pf = PathFinder(arg[0])
                with self.assertRaises(ValueError):
                    pf.add_dirs(arg[1])
        

    @patch('file_navigator.pathfinder.os.path.isdir')     
    def test_del_dirs_iterable_arg(self, mock_isdir):
        directories =  {"/folder1": True,
                        "/folder2/subfolderA": False,
                        "/folder3/subfolderB/subsubfolderC": True}
        
        expected = {"/folder2/subfolderA": False,
                    "/folder3/subfolderB/subsubfolderC": True}
        
        pf = PathFinder()
        pf.add_dirs(directories)
        pf.del_dirs(["/folder1"])
        
        
        self.assertEqual(expected, pf.directories)

    @patch('file_navigator.pathfinder.os.path.isdir')     
    def test_del_dirs_bad_arg(self, mock_isdir):
        directories =  {"/folder1": True,
                        "/folder2/subfolderA": False,
                        "/folder3/subfolderB/subsubfolderC": True}        
        pf = PathFinder()
        pf.add_dirs(directories)
        self.assertRaises(KeyError, pf.del_dirs, '/folder4')

    @patch('file_navigator.pathfinder.os.path.isdir')        
    def test_find_arg_validation(self, mock_isdir):
        directory = r"C:\mock_directory"
        mock_isdir.side_effect = lambda path: True if path == directory else False
        pf = PathFinder({directory: False})
        
        args = [
            (1, 'x'),
            ('*', 2,),
            ('H1', 'c', 3, 'isin'),
            ('H1', 'c', 'abcd', 'isin'),
            ('eurgbp|gbpeur|EURGBP_H4', 'txt|csv', 'regex', 'abcd')
            ]
        expected = {
           (1, 'x'): TypeError,
            ('*', 2,):TypeError,
            ('H1', 'c', 3, 'isin'): TypeError,
            ('H1', 'c', 'abcd', 'isin'): ValueError,
            ('eurgbp|gbpeur|EURGBP_H4', 'txt|csv', 'regex', 'abcd'): ValueError
                }


        for arg in args:
            with self.subTest(arg = arg):
                with self.assertRaises(expected[arg]):
                    pf.find(*arg)

    @patch('file_navigator.pathfinder.os.walk')
    @patch('file_navigator.pathfinder.os.path.isdir')
    @patch('file_navigator.pathfinder.os.path.isfile')
    def test_find_nested_dir(self, mock_isfile, mock_isdir, mock_walk):
        directory = r"C:\mock_directory"
        mock_isdir.side_effect = lambda path: True if path == directory else False
        
        mock_files = ['Forex.xlsx','EURGBP_H4.csv','EURGBP_M5.csv',
                      'EURJPY_H1.csv','EURJPY_M30.csv', 'EURUSD_M5.csv',
                      'audcad.txt','audchf.txt','audusd.txt',
                      'cadaud.txt','cadeur.txt','cadjpy.txt',
                      'cadpln.txt','cadusd.txt','chfgbp.txt',
                      'chfpln.txt','chfusd.txt','euraud.txt',
                      'eurchf.txt','eurgbp.txt','eurjpy.txt',
                      'eurpln.txt','gbpcad.txt','gbpchf.txt',
                      'gbpeur.txt','gbpjpy.txt','gbppln.txt',
                      'jpyaud.txt','jpypln.txt','nzdusd.txt',
                      'usdcad.txt','usdchf.txt','usdjpy.txt',
                      'usdpln.txt','xaggbp.txt','xauchf.txt',
                      'xaueur.txt','xaugbp.txt']

            
        mock_walk.return_value = sorted([
            [directory, ['EUR', 'CURR'], [f]]  if 'xlsx' in f 
            else ['\\'.join([directory, 'EUR']), [], [f]] if 'csv' in f 
            else ['\\'.join([directory, 'CURR']), [], [f]] 
            for f in mock_files
            ])
        mock_isfile.side_effect = lambda path: True if path.split('\\')[-1] in mock_files else False
        
        pf = PathFinder({directory: True})
        args = [
            ('.*_.*|F.*', '.x.*|.c.*', 'regex', 'regex'),
            ('*', '*', 'glob', 'glob'),
            ('H1', 'c', 'isin', 'isin'),
            ('eurgbp|gbpeur|EURGBP_H4', 'txt|csv', 'regex', 'regex'),
            ('Forex', '.xlsx', 'eq', 'eq')
            ]
        expected = {
            ('.*_.*|F.*', '.x.*|.c.*', 'regex', 'regex'): [
                'EURGBP_H4.csv', 'EURGBP_M5.csv','EURJPY_H1.csv',
                'EURJPY_M30.csv','EURUSD_M5.csv', 'Forex.xlsx'
                ],
            ('*', '*', 'glob', 'glob'): mock_files,
            ('H1', 'c', 'isin', 'isin'): ['EURJPY_H1.csv'],
            ('eurgbp|gbpeur|EURGBP_H4', 'txt|csv', 'regex', 'regex'): [
                'EURGBP_H4.csv', 'eurgbp.txt', 'gbpeur.txt'
                ],
            ('Forex', '.xlsx', 'eq', 'eq'): ['Forex.xlsx']
                }
        
        for arg in args:
            with self.subTest(arg = arg):
                result = [p[0] for p in pf.find(*arg).paths]
                self.assertCountEqual(result, expected[arg])
       

    @patch('file_navigator.pathfinder.os.scandir')
    @patch('file_navigator.pathfinder.os.path.isdir')
    @patch('file_navigator.pathfinder.os.path.isfile')
    def test_find_flat_dir(self, mock_isfile, mock_isdir, mock_scandir):
        directory = r"C:\mock_directory"
        mock_isdir.side_effect = lambda path: True if path == directory else False
        
        mock_files = ['EURGBP_H4.csv','EURGBP_M5.csv','EURJPY_H1.csv',
                      'EURJPY_M30.csv', 'EURUSD_M5.csv','Forex.xlsx',
                      'audcad.txt','audchf.txt','audusd.txt',
                      'cadaud.txt','cadeur.txt','cadjpy.txt',
                      'cadpln.txt','cadusd.txt','chfgbp.txt',
                      'chfpln.txt','chfusd.txt','euraud.txt',
                      'eurchf.txt','eurgbp.txt','eurjpy.txt',
                      'eurpln.txt','gbpcad.txt','gbpchf.txt',
                      'gbpeur.txt','gbpjpy.txt','gbppln.txt',
                      'jpyaud.txt','jpypln.txt','nzdusd.txt',
                      'usdcad.txt','usdchf.txt','usdjpy.txt',
                      'usdpln.txt','xaggbp.txt','xauchf.txt',
                      'xaueur.txt','xaugbp.txt']
        mock_scandir.return_value = mock_files
        mock_isfile.side_effect = lambda path: True if path.split('\\')[-1] in mock_files else False
        
        pf = PathFinder({directory: False})
        args = [
            ('.*_.*|F.*', '.x.*|.c.*', 'regex', 'regex'),
            ('*', '*', 'glob', 'glob'),
            ('H1', 'c', 'isin', 'isin'),
            ('eurgbp|gbpeur|EURGBP_H4', 'txt|csv', 'regex', 'regex'),
            ('Forex', '.xlsx', 'eq', 'eq')
            ]
        expected = {
            ('.*_.*|F.*', '.x.*|.c.*', 'regex', 'regex'): [
                'EURGBP_H4.csv', 'EURGBP_M5.csv','EURJPY_H1.csv',
                'EURJPY_M30.csv','EURUSD_M5.csv', 'Forex.xlsx'
                ],
            ('*', '*', 'glob', 'glob'): mock_files,
            ('H1', 'c', 'isin', 'isin'): ['EURJPY_H1.csv'],
            ('eurgbp|gbpeur|EURGBP_H4', 'txt|csv', 'regex', 'regex'): [
                'EURGBP_H4.csv', 'eurgbp.txt', 'gbpeur.txt'
                ],
            ('Forex', '.xlsx', 'eq', 'eq'): ['Forex.xlsx']
                }
        
        for arg in args:
            with self.subTest(arg = arg):
                result = [p[0] for p in pf.find(*arg).paths]
                self.assertCountEqual(result, expected[arg])
    
    def test_empty_find(self):
        pf = PathFinder()
        with self.assertRaises(ValueError):
            pf.find('Forex', '.xlsx', 'eq', 'eq')
        
    
def suite():
    suite = unittest.TestSuite()
    suite.addTest(TestPathFinder('test_empty_init'))
    suite.addTest(TestPathFinder('test_init_good_arg'))
    suite.addTest(TestPathFinder('test_init_bad_arg_1'))
    suite.addTest(TestPathFinder('test_init_bad_arg_2'))
    suite.addTest(TestPathFinder('test_init_bad_arg_3'))
    suite.addTest(TestPathFinder('test_init_bad_arg_4'))
    suite.addTest(TestPathFinder('test_init_bad_arg_5'))
    suite.addTest(TestPathFinder('test_init_bad_arg_6'))
    suite.addTest(TestPathFinder('test_add_dirs'))
    suite.addTest(TestPathFinder('test_del_dirs_iterable_arg'))
    suite.addTest(TestPathFinder('test_del_dirs_bad_arg'))
    suite.addTest(TestPathFinder('test_find_arg_validation'))
    suite.addTest(TestPathFinder('test_find_nested_dir'))
    suite.addTest(TestPathFinder('test_find_flat_dir'))
    suite.addTest(TestPathFinder('test_empty_find'))
    return suite

if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())



import os
from itertools import chain, filterfalse, groupby
from functools import lru_cache, partial
from pathlib import Path
from . import matching
import inspect
from .abc_loader import ABLoader



class _PathManager:
    """
    Private class for file path operations.

    Parameters:
        paths (List[Tuple[str, str]]): Two-element Tuple or List of Tuples,
            containing path-like strings and file names combined with file extensions.

    Attributes:
        paths (List[Tuple[str, str]]): Returns the list of the paths attribute,
            that the class was instantiated with in reversed order (file name is
            the first item in the tuple instead of the file path).
        matching_eng (Type(matching)): Class with the matching functions.

    Methods:
        select_paths (pattern: str, match_type: str): Allows filtering of the file paths
            based on the given pattern and matching function from matching_eng.
            Default match type is equality check: 'eq'.
            The method returns a new instance of _PathManager.
        groupby (by: str, pattern: str, match_type: str): Allows grouping
            the paths based on the following path elements: extension, name, and path.
            Each of the grouping keys also supports matching by pattern
            and matching functions from matching_eng.
            The method returns a dictionary with the group key and a new instance of _PathManager
            instantiated with group values.
        load (Loader, **kwargs): Loads data from the file specified by a single path.
        path (pattern: str, match_type: str, it): Key function for grouping paths
            by file path.
        name (pattern: str, match_type: str, it): Key function for grouping paths
            by file name.
        ext (pattern: str, match_type: str, it): Key function for grouping paths
            by file type.
    """    
    def __init__(self, paths):
        if len(paths) == 0:
            raise ValueError('"paths" parameter is empty.')
        
        if isinstance(paths, tuple):
            self._paths = [paths]
        else:
            self._paths = paths
        self.matching_eng = matching
           
    def __len__(self):
        return len(self._paths)
                  
    def load(self, loader, **kwargs):
        """"
        Loads data from the file specified by a single path-like string.

        This method uses dependency injection to leverage an object following
        abc_Loader.ABLoader interface, to load data from all of the paths
        that the _PathManager was instantiated with.

        Parameters
        ----------
        loader: type[abc_Loader.ABLoader]
            Object that has a load method defined.
        kwargs: dict
            Key-value parameters that are supported by the loader object.

        Returns
        -------
        list
            List of the loaded data objects, e.g., pandas DataFrames.
        """
        if not isinstance(loader, ABLoader):
            raise TypeError("Incorrect Loader type. It must be ABLoader type")
        return [loader.load(os.path.join(*p), **kwargs) for p in self._paths]
    
    def select_paths(self, pattern, match_type = 'eq'):
        """"
        Filters path-like strings based on a specified pattern and creates a new object.

        This method allows filtering file paths based on a given pattern
        that is supported by the types defined in the matching module.

        Parameters
        ----------
        pattern: str
            String that can be matched with a file path.
        match_type: str, default='eq'
            String representing a matching function from the matching module.

        Returns
        -------
        _PathManager
            New instance of _PathManager with filtered paths.
            """
        return self.__class__(
            list(
                filter(
                    lambda p: getattr(self.matching_eng, match_type)(p[0], pattern), 
                    self._paths
                    )
                )
            )
    
    @property
    def paths(self):
        """
        Returns a list with file paths and file names in reversed order.
        """
        return list(tuple(reversed(p)) for p in self._paths)

    def groupby(self, by, pattern = None, match_type = 'eq'):
        """
        Groups paths by a specified part and pattern.

        This function allows grouping paths by a specified part defined by
        a key function (file path, file name, and file type) as well as a specified
        pattern that is supported by the types defined in the matching module.
        The function returns a dictionary with keys defined by the key function
        and values which are new instances of _PathManager.

        Parameters
        ----------
        by: str
            String representing a key function (path, name, ext).
        pattern: str, default=None
            String that can be matched with a file path part.
        match_type: str, default='eq'
            String representing a matching function from the matching module.

        Returns
        -------
        dict
            Dictionary with keys defined by a key function and values which are
            new instances of _PathManager.
        """
        return {
            k:self.__class__(list(g)) for k, g in groupby(
                sorted(
                    self._paths, key= partial(getattr(self, by), pattern, match_type)
                    ),   partial(getattr(self, by), pattern, match_type)
                )
            }
    
    #Sorting functions   
    def path(self, pattern, match_type, it):
        """
        Key function representing file path.

        This function returns a file path part of a single path item.

        Parameters
        ----------
        pattern: str
            String that can be matched with a file path.
        match_type: str
            String representing a matching function from the matching module.

        Returns
        -------
        str
            A file path.
        """
        if pattern is None:
            return it[0]
        else:
            return getattr(self.matching_eng, match_type)(it[0], pattern)

    def name(self, pattern, match_type, it):
        """
        Key function representing file name.

        This function returns a file name part of a single path item.

        Parameters
        ----------
        pattern: str
            String that can be matched with a file name (without file type).
        match_type: str
            String representing a matching function from the matching module.

        Returns
        -------
        str
            A file name (without file type).
        """
        if pattern is None:
            return Path(it[1]).stem
        else:
            return getattr(self.matching_eng, match_type)(Path(it[1]).stem, pattern)


    def ext(self, pattern, match_type, it):
        """
        Key function representing file type.

        This function returns a file type part of a single path item.

        Parameters
        ----------
        pattern: str
            String that can be matched with a file type.
        match_type: str
            String representing a matching function from the matching module.

        Returns
        -------
        str
            A file type.
        """
        if pattern is None:
            return Path(it[1]).suffix
        else:
            return getattr(self.matching_eng, match_type)(Path(it[1]).suffix, pattern)
        

      
        

class PathFinder:
    """
    Main class for navigating through directories and finding files.

    Class for navigating through directories and finding matching file paths,
    supporting pattern matching as defined in the matching module, for file name 
    and file type respectively. Can be instantiated empty or with a dictionary
    with directories to be scanned.

    Attributes:
        directories (Dict): Empty dictionary to which directory path and flag
            for flat or deep scan key, value pairs will be added.
        matching_eng (Type(matching)): Class with the matching functions.
        pm (Type(_PathManager)): Private class for file path operations.

    Parameters:
        init_dirs (Dict[str: bool], default=None): Dictionary with path-like 
            string as key and bool value.

    Methods:
        add_dir (directory: str, traverse_subdirs: bool, default=False): Method
            for adding single key, value pair of directory path with a bool flag 
            indicating deep or flat scan.
        del_dir (directory: str): Method for removing single key, value pair of
            directory path with traverse_subdirs flag.
        add_dirs (directories: Dict[str: bool]): Method for appending
            a collection of directory path and traverse_subdirs flag pairs.
        del_dirs (directories: str | List[str]): Method for deleting a collection 
            of directory path and traverse_subdirs flag pairs.
        find (name: str, ext: str, name_type: str[default='eq'], ext_type: str[default='eq']):
            Method for iterating through all of the directories collection and matching 
            files based on defined file name and file type patterns, supported 
            by the matching_eng.
    """    
    def __init__(self, init_dirs = None):
        self.directories = {}
        self.matching_eng = matching
        self.pm = _PathManager

        
        if init_dirs is not None:
            self.add_dirs(init_dirs)
            
    def add_dir(self, directory, traverse_subdirs = False):
        """
        Function for adding a directory entry.
        
        This function allows adding a single directory entry, validating whether
        the passed key is a valid directory.
        
        Parameters
        ----------
        directory: str
            Path-like string pointing to an existing directory.
        traverse_subdirs: bool, default=False
            Flag indicating whether all subdirectories of the passed directory should 
            be iterated over.
            
        Returns
        -------
        None        
        """
        if os.path.isdir(directory):
            if any(self._overlap(directory, traverse_subdirs)):
               raise ValueError(f"{directory} can't be added due to conflicting "\
                                "parent - child relationship with already added "\
                                f"directories: {', '.join(self._overlap(directory, traverse_subdirs))}")
            self.directories[directory] = traverse_subdirs
        else:
            raise ValueError("Specified directory does not exist")
            
        if not (isinstance(traverse_subdirs, (bool, int)) and int(traverse_subdirs) <= 1):
            raise TypeError("'traverse_subdirs' argument must be bool or int: (0,1)")
    
    @lru_cache(maxsize=None)        
    def _overlap(self, directory, traverse_subdirs):
        return [d for d, t_s in self.directories.items() if (
            (
                (d in directory and t_s ) or (directory in d and traverse_subdirs)
                )
            and len(d) != len(directory)
            )]

            
    def del_dir(self, directory):
        """
        Function for removing a directory entry.
        
        This function allows the removal of a single directory entry. If the specified
        directory is not part of the directories collection, a KeyError is raised.
        
        Parameters
        ----------
        directory: str
            Path-like string pointing to a directory in the directories collection.
            
        Returns
        -------
        None
        """
        del self.directories[directory]
        
    def add_dirs(self, directories):
        """
        Function for adding multiple directory entries.
        
        This function allows adding multiple directory entries passed as a dictionary, 
        iteratively calling the add_dir method.
        
        Parameters
        ----------
        directories: Dict[str: bool]
            Dictionary containing Path-like strings pointing to existing 
            directories with a traverse_subdirs flag.
            
        Returns
        -------
        None
        """
        for k, v in directories.items():
            self.add_dir(k, v)
    
    def del_dirs(self, directories):
        """
        Function for removing multiple directory entries.
        
        This function allows removing multiple directory entries passed in a list, 
        iteratively calling the del_dir method.
        
        Parameters
        ----------
        directories: str | List[str]
            List containing path-like strings or a single string pointing to existing 
            directory entries.
            
        Returns
        -------
        None
        """
        if not isinstance(directories, (list, str)):
            raise TypeError('"directories" argument must be a list os strings or a string')
        for d in directories:
            self.del_dir(d)
            
    @lru_cache(maxsize=None)
    def _resolve_ext(self, string):
        """
        Private function for standardizing file type strings.

        Parameters
        ----------
        string: str
            File extension string to be standardized.

        Returns
        -------
        str
            Standardized file extension without the dot prefix.
        """        
        if '.' in string:
            return string.replace('.', '')
        else:
            return string

          
    def _traverse_subdir(self, directory, name, ext, name_type, ext_type):
        """
        Private function for nested directory iteration and file matching.
        
        This function iterates through a single directory, including all subdirectories,
        trying to match all files based on the specified name and file type patterns. 
        
        Parameters
        ----------
        directory: str
            Path-like string pointing to an existing directory.
        name: str
            File name pattern to be matched.
        ext: str
            File type (file extension) pattern to be matched.
        name_type: str
            String representing a function in matching_eng for matching 
            the file name pattern.
        ext_type: str
            String representing a function in matching_eng for matching 
            the file type (extension) pattern.
        
        Returns
        -------
        Generator[Tuple[root[str], file[str]]]
            Generator containing a 2-element tuple with the root directory 
            and the matching file.
        """
        return ((root, file) for root, _, files in os.walk(directory) 
                for file in files 
                if os.path.isfile(os.path.join(root, file))
                and getattr(self.matching_eng, ext_type)(self._resolve_ext(Path(file).suffix), ext)
                and getattr(self.matching_eng, name_type)(Path(file).stem, name))


                
    def _traverse_dir(self, directory, name, ext, name_type, ext_type):
        """
        Private function for flat directory iteration and file matching.
        
        This function iterates through a single directory, trying to match 
        all files based on the specified name and file type patterns. 
        
        Parameters
        ----------
        directory: str
            Path-like string pointing to an existing directory.
        name: str
            File name pattern to be matched.
        ext: str
            File type (file extension) pattern to be matched.
        name_type: str
            String representing a function in matching_eng for matching 
            the file name pattern.
        ext_type: str
            String representing a function in matching_eng for matching 
            the file type (extension) pattern.
        
        Returns
        -------
        Generator[Tuple[directory[str], file[str]]]
            Generator containing a 2-element tuple with the directory 
            and the matching file.
        """
        return ((directory, file)  for file in os.scandir(directory) 
                if os.path.isfile(os.path.join(directory, file))
                and getattr(self.matching_eng, ext_type)(self._resolve_ext(Path(file).suffix), ext)
                and getattr(self.matching_eng, name_type)(Path(file).stem, name))

    @lru_cache(maxsize=None)
    def _get_obj_func(self, obj):
        """
        Private function to dynamically return a joined string of a list 
        containing all available functions in a specified object.
        
        Parameters
        ----------
        obj: Any
            The object whose functions will be retrieved.
        
        Returns
        -------
        str
            String of all object methods joined by commas.
        """
        return ', '.join(
            i[0] for i in inspect.getmembers(obj, predicate=inspect.isfunction)
            )

    @lru_cache(maxsize=None)
    def find(self, name, ext, name_type = 'eq', ext_type = 'eq'):
        """
        Function for finding files in defined directories.
        
        This function iterates through all directories in the directories attribute, 
        returning a new instance of the _PathManager class instantiated with all 
        unique matching files and paths pointing to them. Files are matched by both 
        file name and type patterns that are supported by matching_eng.
        
        Parameters
        ----------
        name: str
            File name pattern to be matched.
        ext: str
            File type (file extension) pattern to be matched.
        name_type: str, default='eq'
            String representing a function in matching_eng for matching 
            the file name pattern.
        ext_type: str, default='eq'
            String representing a function in matching_eng for matching 
            the file type (extension) pattern.
        
        Returns
        -------
        Type[_PathManager]
            New instance of the _PathManager class.
        """
        if not isinstance(name, str):
            raise TypeError('"name" argument must be string type')     
            
        if not isinstance(ext, str):
            raise TypeError('"ext" argument must be string type')     
            
        if not hasattr(matching, name_type):
            raise ValueError(f'"name_type" argument must be one of {self._get_obj_func(matching)}')
            
        if not hasattr(matching, ext_type):
            raise ValueError(f'"name_type" argument must be one of {self._get_obj_func(matching)}')
            
        if len(self.directories) == 0:
            raise ValueError('There are no dictionaries to be searched.')


        return self.pm(
                set(
                    filterfalse(
                        lambda path: path is False, 
                        chain.from_iterable(
                            [self._traverse_subdir(directory, 
                                                   name, self._resolve_ext(ext),  
                                                   name_type, ext_type) 
                             if traverse_subdirs 
                             else self._traverse_dir(directory, 
                                                     name, self._resolve_ext(ext),  
                                                     name_type, ext_type)
                             for directory, traverse_subdirs in self.directories.items()]
                            )
                        )
                    )
                )
            
    
